package ap.terfor.arithconj;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.equations.{EquationConj, NegEquationConj, ReduceWithEqs}
import ap.terfor.inequalities.InEqConj
import ap.terfor.preds.Predicate
import ap.util.{Debug, Logic}

object ArithConj {
  
  val AC = Debug.AC_PROP_CONNECTIVES
  
  def apply(positiveEqs : EquationConj, negativeEqs : NegEquationConj,
            inEqs : InEqConj, order : TermOrder) : ArithConj = {
    if (positiveEqs.isFalse || negativeEqs.isFalse || inEqs.isFalse)
      FALSE
    else if (positiveEqs.isTrue && negativeEqs.isTrue && inEqs.isTrue)
      TRUE
    else
      new ArithConj (positiveEqs, negativeEqs, inEqs, order)
  }
  
  def unapply(conj : ArithConj) : Option[(EquationConj, NegEquationConj, InEqConj)] =
    Some(conj.positiveEqs, conj.negativeEqs, conj.inEqs)
  
  val TRUE : ArithConj = new ArithConj(EquationConj.TRUE, NegEquationConj.TRUE,
                                       InEqConj.TRUE, TermOrder.EMPTY)

  val FALSE : ArithConj = new ArithConj(EquationConj.FALSE, NegEquationConj.TRUE,
                                        InEqConj.TRUE, TermOrder.EMPTY)
                                        
  /**
   * Compute the conjunction of equations, inequations and inequalities.
   */
  def conj(conjs : Iterable[Formula], order : TermOrder) : ArithConj = {
    val posEqs = new ArrayBuffer[EquationConj]
    val negEqs = new ArrayBuffer[NegEquationConj]
    val inEqs = new ArrayBuffer[InEqConj]
    
    for (conj <- conjs) {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPre(AC, order isSortingOf conj)
      //////////////////////////////////////////////////////////////////////////
      
      conj match {
      case conj : EquationConj => posEqs += conj
      case conj : NegEquationConj => negEqs += conj
      case conj : InEqConj => inEqs += conj
      case conj : ArithConj => {
        posEqs += conj.positiveEqs
        negEqs += conj.negativeEqs
        inEqs += conj.inEqs
      }
      }
    }
    
    apply(EquationConj.conj(posEqs, order),
          NegEquationConj.conj(negEqs, order),
          InEqConj.conj(inEqs, order),
          order)
  }

  def conj(f : Formula, order : TermOrder) : ArithConj = conj(List(f), order)

}

/**
 * The class for a conjunction of equations, negated equations and inequalities
 */
class ArithConj private (val positiveEqs : EquationConj,
                         val negativeEqs : NegEquationConj,
                         val inEqs : InEqConj,
                         val order : TermOrder)
                extends Formula with SortedWithOrder[ArithConj] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ArithConj.AC,
                   (positiveEqs isSortedBy order) &&
                   (negativeEqs isSortedBy order) &&
                   (inEqs isSortedBy order) &&
                   !negativeEqs.isFalse && !inEqs.isFalse)
  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : ArithConj = {
    if (isSortedBy(newOrder)) {
      this
    } else {
      ArithConj (positiveEqs sortBy newOrder,
                 negativeEqs sortBy newOrder,
                 inEqs sortBy newOrder,
                 newOrder)
    }
  }
   
  //////////////////////////////////////////////////////////////////////////////

  def isTrue : Boolean = (positiveEqs.isTrue && negativeEqs.isTrue && inEqs.isTrue)

  def isFalse : Boolean = (positiveEqs.isFalse || negativeEqs.isFalse || inEqs.isFalse)

  def size : Int = (positiveEqs.size + negativeEqs.size + inEqs.size)
  
  def isEmpty : Boolean = size == 0
  
  /**
   * Return whether this conjunction actually only is a single literal
   */
  def isLiteral : Boolean = (this.size == 1)

  /**
   * Create the negation of at most one equation
   */
  def negate : ArithConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ArithConj.AC, this.size <= 1)
    ////////////////////////////////////////////////////////////////////////////
    if (!negativeEqs.isEmpty) {
      ArithConj(negativeEqs.negate, NegEquationConj.TRUE, InEqConj.TRUE, order)
    } else if (!positiveEqs.isEmpty) {
      ArithConj(EquationConj.TRUE, positiveEqs.negate, InEqConj.TRUE, order)
    } else if (!inEqs.isEmpty) {
      ArithConj(EquationConj.TRUE, NegEquationConj.TRUE, inEqs.negate, order)
    } else {
      ArithConj.FALSE
    }
  }

  /**
   * Update the positive equations of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updatePositiveEqs(newEqs : EquationConj, newOrder : TermOrder) : ArithConj =
    if (this.positiveEqs == newEqs)
      this
    else
      ArithConj(newEqs, this.negativeEqs, this.inEqs, newOrder)
  
  /**
   * Update the Negative equations of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updateNegativeEqs(newEqs : NegEquationConj, newOrder : TermOrder) : ArithConj =
    if (this.negativeEqs == newEqs)
      this
    else
      ArithConj(this.positiveEqs, newEqs, this.inEqs, newOrder)

  /**
   * Update the inequalities of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updateInEqs(newInEqs : InEqConj, newOrder : TermOrder) : ArithConj =
    if (this.inEqs == newInEqs)
      this
    else
      ArithConj(this.positiveEqs, this.negativeEqs, newInEqs, newOrder)

  //////////////////////////////////////////////////////////////////////////////

  lazy val variables : Set[VariableTerm] =
    positiveEqs.variables ++ negativeEqs.variables ++ inEqs.variables
    
  lazy val constants : Set[ConstantTerm] =
    positiveEqs.constants ++ negativeEqs.constants ++ inEqs.constants
      
  def predicates : Set[Predicate] = Set.empty

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    if (isTrue) {
      "true"
    } else if (isFalse) {
      "false"
    } else {
      val strings = (for (lhs <- positiveEqs.elements) yield ("" + lhs + " = 0")) ++
                    (for (lhs <- negativeEqs.elements) yield ("" + lhs + " != 0")) ++
                    (for (lhs <- inEqs.elements) yield ("" + lhs + " >= 0"))
      if (strings.hasNext)
        strings.reduceLeft((s1 : String, s2 : String) => s1 + " & " + s2)
      else
        throw new Error // should never be reached
    }
  }

  override def equals(that : Any) : Boolean = that match {
    case that : ArithConj => this.positiveEqs == that.positiveEqs &&
                             this.negativeEqs == that.negativeEqs &&
                             this.inEqs == that.inEqs
    case _ => false
  }

  override def hashCode = (positiveEqs.hashCode +
                           726327 * negativeEqs.hashCode +
                           36323 * inEqs.hashCode)
}
