package ap.terfor.conjunctions;

import ap.terfor.arithconj.{ArithConj, ReduceWithAC}
import ap.terfor.equations.EquationConj
import ap.terfor.preds.{PredConj, ReduceWithPredLits}
import ap.terfor.substitutions.Substitution
import ap.util.{Debug, Logic, PlainRange}


object ReduceWithConjunction {
  
  private val AC = Debug.AC_PROPAGATION

  def apply(conj : Conjunction, order : TermOrder) : ReduceWithConjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, (conj isSortedBy order) &&
                    // conjunctions with quantifiers are not supported right now
                    conj.quans.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    
    new ReduceWithConjunction(ReduceWithAC(conj.arithConj, order),
                              ReduceWithPredLits(conj.predConj, order),
                              order)
  }
  
  /**
   * Recursively reduce an arbitrary conjunction
   */
  private def reduceConj(conj : Conjunction,
                         initialReducer : ReduceWithConjunction) : Conjunction =
    if (conj.isTrue)
      conj
    else try {
      val (newArithConj, reducer) =
        initialReducer.passQuantifiers(conj.quans.size) reduce conj.arithConj
    
      val (newPredConj, reducer2) = reducer reduce conj.predConj

      val newNegConjs = conj.negatedConjs.update(for (c <- conj.negatedConjs)
                                                 yield reduceConj(c, reducer2),
                                                 reducer2.order)
       
      val res = constructConj(conj, newArithConj, newPredConj, newNegConjs,
                              reducer2.order)    
      if (newArithConj == res.arithConj) {
        res
      } else {
        // it might be necessary to repeat reduction, because new facts became
        // available
        reduceConj(res, initialReducer)
      }
    } catch {
      case FALSE_EXCEPTION => Conjunction.FALSE
    }

  private def constructConj(oldConj : Conjunction,
                            newArithConj : ArithConj,
                            newPredConj : PredConj,
                            newNegConjs : NegatedConjunctions,
                            order : TermOrder) : Conjunction = 
    if (oldConj.quans.isEmpty) {
      Conjunction(oldConj.quans, newArithConj, newPredConj, newNegConjs, order)
    } else {
      var eliminableVars : Set[Term] = Set()
    
      var i : Int = 0
      while (i < oldConj.quans.size && oldConj.quans(i) == Quantifier.EX) {
        val variable = VariableTerm(i)
        if (!(newNegConjs contains variable))
          eliminableVars = eliminableVars + variable
        i = i + 1
      }
    
      val literals = Conjunction.conj(Array(newArithConj, newPredConj), order)
      val essentialLits = 
        if (eliminableVars.isEmpty)
          literals
        else
          new LiteralEliminator(literals, eliminableVars, order).eliminate
    
      Conjunction(oldConj.quans, essentialLits.arithConj, essentialLits.predConj,
                  newNegConjs, order)
    }
  
}

class ReduceWithConjunction private (private val acReducer : ReduceWithAC,
                                     private val predReducer : ReduceWithPredLits,
                                     private val order : TermOrder) {

  def passQuantifiers(num : Int) : ReduceWithConjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithConjunction.AC, num >= 0)
    ////////////////////////////////////////////////////////////////////////////
    if (num == 0)
      this
    else
      new ReduceWithConjunction(acReducer passQuantifiers num,
                                predReducer passQuantifiers num,
                                order)
  }

  def addArithConj(ac : ArithConj) : ReduceWithConjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithConjunction.AC, ac isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////

    if (ac.isEmpty)
      this
    else
      new ReduceWithConjunction(acReducer addArithConj ac, predReducer, order)
  }

  def apply(conj : Conjunction) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithConjunction.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
      
    val res = ReduceWithConjunction.reduceConj(conj, this)
    ////////////////////////////////////////////////////////////////////////////
    // we demand that the reducer is a projection (repeated application does not
    // change the result anymore)
    Debug.assertPost(ReduceWithConjunction.AC,
                     ReduceWithConjunction.reduceConj(res, this) == res)
    ////////////////////////////////////////////////////////////////////////////
    res
  }

  def apply(conjs : NegatedConjunctions) : NegatedConjunctions = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithConjunction.AC, conjs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////

    val res = conjs.update(for (c <- conjs)
                           yield (ReduceWithConjunction.reduceConj(c, this)),
                           order)
    ////////////////////////////////////////////////////////////////////////////
    // we demand that the reducer is a projection (repeated application does not
    // change the result anymore)
    Debug.assertPost(ReduceWithConjunction.AC,
                     Logic.forall(for (c <- res.elements) yield (this(c) == c)))
    ////////////////////////////////////////////////////////////////////////////
    res
  }

  def apply(conj : EquationConj) : EquationConj = acReducer(conj)
  def apply(conj : ArithConj) : ArithConj = acReducer(conj)
  
  private def replaceAC(newAC : ReduceWithAC) : ReduceWithConjunction =
    if (newAC == this.acReducer)
      this
    else
      new ReduceWithConjunction(newAC, predReducer, order)
  
  private def replacePred(newPred : ReduceWithPredLits) : ReduceWithConjunction =
    if (newPred == this.predReducer)
      this
    else
      new ReduceWithConjunction(acReducer, newPred, order)
  
  private def reduce(ac : ArithConj) : (ArithConj, ReduceWithConjunction) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithConjunction.AC, ac isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val (newArithConj, newACReducer) = acReducer reduceAndAdd ac
    if (newArithConj.isFalse) throw FALSE_EXCEPTION
    (newArithConj, this replaceAC newACReducer)
  }
  
  private def reduce(conj : PredConj) : (PredConj, ReduceWithConjunction) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithConjunction.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redConj = predReducer(acReducer(conj))
    if (redConj.isFalse) throw FALSE_EXCEPTION
    (redConj, this replacePred (predReducer addLits redConj))
  }
  
}

private object FALSE_EXCEPTION extends Exception

private class LiteralEliminator(literals : Conjunction,
                                uniVariables : Set[Term],
                                order : TermOrder)
              extends ConjunctEliminator(literals, uniVariables, order) {
  
  protected def nonUniversalElimination(f : Conjunction) =
    throw new UnsupportedOperationException
  
  protected def universalElimination(eliminatedConstant : ConstantTerm,
                  witness : (Substitution, TermOrder) => Substitution) : unit = {
    // nothing ... we do not need any witness information at this point
  }

  protected def eliminationCandidates(literals : Conjunction) : Iterator[Term] =
    uniVariables.elements
  
}